#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
REGISTER:
---------

Caller script. Designed to call all other functions.

'''
import argparse
import logging
from os import getenv

from hdx.facades.keyword_arguments import facade
from hdx.utilities.database import Database
from hdx.utilities.dictandlist import args_to_dict
from hdx.utilities.easy_logging import setup_logging
from hdx.utilities.path import script_dir_plus_file
from hdx.utilities.useragent import UserAgent

from hdx.freshness.datafreshness import DataFreshness
from hdx.freshness.version import get_freshness_version

setup_logging()
logger = logging.getLogger(__name__)


def main(db_url, db_params, do_touch, save, **ignore):
    logger.info('> Data freshness %s' % get_freshness_version())
    if db_params:
        params = args_to_dict(db_params)
    elif db_url:
        params = Database.get_params_from_sqlalchemy_url(db_url)
    else:
        params = {'driver': 'sqlite', 'database': 'freshness.db'}
    logger.info('> Database parameters: %s' % params)
    with Database(**params) as session:
        testsession = None
        if save:
            testsession = Database.get_session('sqlite:///test_serialize.db')
        freshness = DataFreshness(session=session, testsession=testsession, do_touch=do_touch)
        freshness.spread_datasets()
        freshness.add_new_run()
        datasets_to_check, resources_to_check = freshness.process_datasets()
        results, hash_results = freshness.check_urls(resources_to_check, UserAgent.get())
        datasets_lastmodified = freshness.process_results(results, hash_results)
        freshness.update_dataset_latest_of_modifieds(datasets_to_check, datasets_lastmodified)
        freshness.output_counts()
        if testsession:
            testsession.close()
    logger.info('Freshness completed!')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Data Freshness')
    parser.add_argument('-hk', '--hdx_key', default=None, help='HDX api key')
    parser.add_argument('-ua', '--user_agent', default=None, help='user agent')
    parser.add_argument('-pp', '--preprefix', default=None, help='preprefix')
    parser.add_argument('-hs', '--hdx_site', default=None, help='HDX site to use')
    parser.add_argument('-db', '--db_url', default=None, help='Database connection string')
    parser.add_argument('-dp', '--db_params', default=None, help='Database connection parameters. Overrides --db_url.')
    parser.add_argument('-dt', '--donttouch', default=False, action='store_true', help="Don't touch datasets")
    parser.add_argument('-s', '--save', default=False, action='store_true', help='Save state for testing')
    args = parser.parse_args()
    hdx_key = args.hdx_key
    if hdx_key is None:
        hdx_key = getenv('HDX_KEY')
    user_agent = args.user_agent
    if user_agent is None:
        user_agent = getenv('USER_AGENT')
        if user_agent is None:
            user_agent = 'freshness'
    preprefix = args.preprefix
    if preprefix is None:
        preprefix = getenv('PREPREFIX')
    hdx_site = args.hdx_site
    if hdx_site is None:
        hdx_site = getenv('HDX_SITE', 'prod')
    db_url = args.db_url
    if db_url is None:
        db_url = getenv('DB_URL')
    if db_url and '://' not in db_url:
        db_url = 'postgresql://%s' % db_url
    project_config_yaml = script_dir_plus_file('project_configuration.yml', main)
    facade(main, hdx_key=hdx_key, user_agent=user_agent, preprefix=preprefix, hdx_site=hdx_site,
           project_config_yaml=project_config_yaml, db_url=db_url, db_params=args.db_params,
           do_touch=not args.donttouch, save=args.save)



# -*- coding: utf-8 -*-
from hdx.utilities.database import Base
from sqlalchemy import Column, DateTime, Boolean, String


class DBTestHashResult(Base):
    id = Column(String, primary_key=True)
    url = Column(String, nullable=False)
    err = Column(String)
    http_last_modified = Column(DateTime)
    hash = Column(String)
    force_hash = Column(Boolean, nullable=False)

    def __repr__(self):
        output = '<TestHashResult(id=%s, url=%s, ' % (self.id, self.url)
        output += 'err=%s\n' % self.err
        output += 'http_last_modified=%s, hash=%s,' % (str(self.http_last_modified), self.hash)
        output += 'force_hash=%s)>' % str(self.force_hash)
        return output



# -*- coding: utf-8 -*-
from hdx.utilities.database import Base
from sqlalchemy import Column, DateTime


class DBTestDate(Base):
    test_date = Column(DateTime, primary_key=True)

    def __repr__(self):
        return '<Test date=%s>' % str(self.test_date)



# -*- coding: utf-8 -*-
from hdx.utilities.database import Base
from sqlalchemy import Column, String
from sqlalchemy import ForeignKey

from hdx.freshness.testdata.dbtestdataset import DBTestDataset


class DBTestResource(Base):
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    format = Column(String, nullable=False)
    dataset_id = Column(String, ForeignKey(DBTestDataset.id), nullable=False)
    url = Column(String, nullable=False)
    revision_last_updated = Column(String, nullable=False)
    last_modified = Column(String, nullable=False)

    def __repr__(self):
        output = '<Resource(run number=%d, id=%s,\n' % (self.run_number, self.id)
        output += 'format=%s, dataset id=%s,\n' % (self.format, self.dataset_id)
        output += 'url=%s,\n' % self.url
        output += 'revision last updated=%s, ' % str(self.revision_last_updated)
        output += 'last_modified=%s)>' % str(self.last_modified)
        return output



# -*- coding: utf-8 -*-
from hdx.utilities.database import Base
from sqlalchemy import Column, Boolean, String


class DBTestDataset(Base):
    id = Column(String, primary_key=True)
    organization_id = Column(String, nullable=False)
    organization_name = Column(String, nullable=False)
    organization_title = Column(String, nullable=False)
    dataset_name = Column(String, nullable=False)
    dataset_title = Column(String, nullable=False)
    dataset_private = Column(Boolean, nullable=False)
    dataset_maintainer = Column(String)
    dataset_date = Column(String)
    update_frequency = Column(String)
    review_date = Column(String)
    last_modified = Column(String, nullable=False)
    updated_by_script = Column(String)
    metadata_modified = Column(String, nullable=False)
    is_requestdata_type = Column(Boolean)
    dataset_location = Column(String)

    def __repr__(self):
        output = '<Dataset(id=%s, name=%s' % (self.id, self.dataset_name)
        output += 'organisation name=%s,\n' % self.organization_name
        output += 'dataset date=%s, update frequency=%s, ' % (self.dataset_date, self.update_frequency)
        output += 'review_date=%s, last_modified=%s,' % (str(self.review_date), str(self.last_modified))
        output += 'updated_by_script=%s, metadata_modified=%s)>' % (
        str(self.updated_by_script), str(self.metadata_modified))
        return output



# -*- coding: utf-8 -*-
'''
Data freshness:
--------------

Calculate freshness for all datasets in HDX.

'''
import datetime
import logging
import re
from parser import ParserError
from urllib.parse import urlparse

from dateutil import parser
from hdx.data.dataset import Dataset
from hdx.data.hdxobject import HDXError
from hdx.data.resource import Resource
from hdx.hdx_configuration import Configuration
from hdx.utilities.dictandlist import dict_of_lists_add, list_distribute_contents
from sqlalchemy import exists, and_
from sqlalchemy.orm.exc import NoResultFound

from hdx.freshness.database.dbdataset import DBDataset
from hdx.freshness.database.dbinfodataset import DBInfoDataset
from hdx.freshness.database.dborganization import DBOrganization
from hdx.freshness.database.dbresource import DBResource
from hdx.freshness.database.dbrun import DBRun
from hdx.freshness.retrieval import retrieve
from hdx.freshness.testdata.serialize import serialize_datasets, serialize_now, serialize_results, serialize_hashresults

logger = logging.getLogger(__name__)

default_no_urls_to_check = 1000


class DataFreshness:
    bracketed_date = re.compile(r'\((.*)\)')

    def __init__(self, session=None, testsession=None, datasets=None, now=None, do_touch=False):
        ''''''
        self.session = session
        self.urls_to_check_count = 0
        self.never_update = 0
        self.live_update = 0
        self.adhoc_update = 0
        self.dataset_what_updated = dict()
        self.resource_what_updated = dict()
        self.resource_last_modified_count = 0
        self.do_touch = do_touch

        self.url_internal = 'data.humdata.org'

        self.aging = dict()
        for key, value in Configuration.read()['aging'].items():
            period = int(key)
            aging_period = dict()
            for status in value:
                nodays = value[status]
                aging_period[status] = datetime.timedelta(days=nodays)
            self.aging[period] = aging_period
        self.aging_statuses = {0: '0: Fresh', 1: '1: Due', 2: '2: Overdue', 3: '3: Delinquent',
                               None: 'Freshness Unavailable'}
        self.testsession = testsession
        if datasets is None:  # pragma: no cover
            Configuration.read().set_read_only(True)  # so that we only get public datasets
            logger.info('Retrieving all datasets from HDX')
            self.datasets = Dataset.get_all_datasets()
            Configuration.read().set_read_only(False)
            if self.testsession:
                serialize_datasets(self.testsession, self.datasets)
        else:
            self.datasets = datasets
        if now is None:  # pragma: no cover
            self.now = datetime.datetime.utcnow()
            if self.testsession:
                serialize_now(self.testsession, self.now)
        else:
            self.now = now
        self.previous_run_number = self.session.query(DBRun.run_number).distinct().order_by(
            DBRun.run_number.desc()).first()
        if self.previous_run_number is not None:
            self.previous_run_number = self.previous_run_number[0]
            self.run_number = self.previous_run_number + 1
            no_resources = self.no_resources_force_hash()
            if no_resources:
                self.no_urls_to_check = int((no_resources / 30) + 1)
            else:
                self.no_urls_to_check = default_no_urls_to_check
        else:
            self.previous_run_number = None
            self.run_number = 0
            self.no_urls_to_check = default_no_urls_to_check

        logger.info('Will force hash %d resources' % self.no_urls_to_check)

    def no_resources_force_hash(self):
        columns = [DBResource.id, DBDataset.updated_by_script]
        filters = [DBResource.dataset_id == DBDataset.id, DBResource.run_number == self.previous_run_number,
                   DBDataset.run_number == self.previous_run_number,
                   DBResource.url.notlike('%{}%'.format(self.url_internal))]
        query = self.session.query(*columns).filter(and_(*filters))
        noscriptupdate = 0
        noresources = 0
        for result in query:
            updated_by_script = result[1]
            if updated_by_script is not None:
                noscriptupdate += 1
                continue
            noresources += 1
        if noscriptupdate == 0:
            return None
        return noresources

    def spread_datasets(self):
        self.datasets = list_distribute_contents(self.datasets, lambda x: x['organization']['name'])

    def add_new_run(self):
        dbrun = DBRun(run_number=self.run_number, run_date=self.now)
        self.session.add(dbrun)
        self.session.commit()

    def process_datasets(self, forced_hash_ids=None):
        resources_to_check = list()
        datasets_to_check = dict()
        logger.info('Processing datasets')
        for dataset in self.datasets:
            resources = dataset.get_resources()
            if dataset.is_requestable():  # ignore requestable
                continue
            dataset_id = dataset['id']
            dict_of_lists_add(self.dataset_what_updated, 'total', dataset_id)
            organization_id = dataset['organization']['id']
            organization_name = dataset['organization']['name']
            organization_title = dataset['organization']['title']
            try:
                dborganization = self.session.query(DBOrganization).filter_by(id=organization_id).one()
                dborganization.name = organization_name
                dborganization.title = organization_title
            except NoResultFound:
                dborganization = DBOrganization(name=organization_name, id=organization_id, title=organization_title)
                self.session.add(dborganization)
            dataset_name = dataset['name']
            dataset_title = dataset['title']
            dataset_private = dataset['private']
            dataset_maintainer = dataset['maintainer']
            dataset_location = ','.join([x['name'] for x in dataset['groups']])
            try:
                dbinfodataset = self.session.query(DBInfoDataset).filter_by(id=dataset_id).one()
                dbinfodataset.name = dataset_name
                dbinfodataset.title = dataset_title
                dbinfodataset.private = dataset_private
                dbinfodataset.organization_id = organization_id
                dbinfodataset.maintainer = dataset_maintainer
                dbinfodataset.location = dataset_location
            except NoResultFound:
                dbinfodataset = DBInfoDataset(name=dataset_name, id=dataset_id, title=dataset_title,
                                              private=dataset_private, organization_id=organization_id,
                                              maintainer=dataset_maintainer, location=dataset_location)
                self.session.add(dbinfodataset)
            try:
                previous_dbdataset = self.session.query(DBDataset).filter_by(run_number=self.previous_run_number,
                                                                             id=dataset_id).one()
            except NoResultFound:
                previous_dbdataset = None
            dont_hash_script_update = False
            update_frequency = dataset.get('data_update_frequency')
            updated_by_script = dataset.get('updated_by_script')
            if update_frequency is not None:
                update_frequency = int(update_frequency)
                if updated_by_script:
                    if 'freshness_ignore' in updated_by_script:
                        updated_by_script = None
                    else:
                        match = self.bracketed_date.search(updated_by_script)
                        if match is None:
                            updated_by_script = None
                        else:
                            try:
                                updated_by_script = parser.parse(match.group(1), ignoretz=True)
                                dont_hash_script_update = True
                            except ParserError:
                                updated_by_script = None
            dataset_resources, last_resource_updated, last_resource_modified = \
                self.process_resources(dataset_id, previous_dbdataset, resources, dont_hash_script_update,
                                       forced_hash_ids=forced_hash_ids)
            dataset_date = dataset.get('dataset_date')
            metadata_modified = parser.parse(dataset['metadata_modified'], ignoretz=True)
            if 'last_modified' in dataset:
                last_modified = parser.parse(dataset['last_modified'], ignoretz=True)
            else:
                last_modified = datetime.datetime(1970, 1, 1, 0, 0)
            if len(resources) == 0 and last_resource_updated is None:
                last_resource_updated = 'NO RESOURCES'
                last_resource_modified = datetime.datetime(1970, 1, 1, 0, 0)
                error = True
                what_updated = 'no resources'
            else:
                error = False
                what_updated = 'firstrun'
            review_date = dataset.get('review_date')
            if review_date is None:
                latest_of_modifieds = last_modified
            else:
                review_date = parser.parse(review_date, ignoretz=True)
                if review_date > last_modified:
                    latest_of_modifieds = review_date
                else:
                    latest_of_modifieds = last_modified
            if updated_by_script and updated_by_script > latest_of_modifieds:
                latest_of_modifieds = updated_by_script
            fresh = None
            if update_frequency is not None and not error:
                if update_frequency == 0:
                    fresh = 0
                    self.live_update += 1
                elif update_frequency == -1:
                    fresh = 0
                    self.never_update += 1
                elif update_frequency == -2:
                    fresh = 0
                    self.adhoc_update += 1
                else:
                    fresh = self.calculate_aging(latest_of_modifieds, update_frequency)
            dbdataset = DBDataset(run_number=self.run_number, id=dataset_id,
                                  dataset_date=dataset_date, update_frequency=update_frequency,
                                  review_date=review_date, last_modified=last_modified,
                                  metadata_modified=metadata_modified, updated_by_script=updated_by_script,
                                  latest_of_modifieds=latest_of_modifieds, what_updated=what_updated,
                                  last_resource_updated=last_resource_updated,
                                  last_resource_modified=last_resource_modified, fresh=fresh, error=error)
            if previous_dbdataset is not None and not error:
                dbdataset.what_updated = self.add_what_updated(dbdataset.what_updated, 'nothing')
                if last_modified > previous_dbdataset.last_modified:  # filestore update would cause this
                    dbdataset.what_updated = self.add_what_updated(dbdataset.what_updated, 'filestore')
                else:
                    dbdataset.last_modified = previous_dbdataset.last_modified
                if previous_dbdataset.review_date is None:
                    if review_date is not None:
                        dbdataset.what_updated = self.add_what_updated(dbdataset.what_updated, 'review date')
                else:
                    if review_date is not None and review_date > previous_dbdataset.review_date:  # someone clicked the review button
                        dbdataset.what_updated = self.add_what_updated(dbdataset.what_updated, 'review date')
                    else:
                        dbdataset.review_date = previous_dbdataset.review_date
                if updated_by_script and (
                        previous_dbdataset.updated_by_script is None or updated_by_script > previous_dbdataset.updated_by_script):  # new script update of datasets
                    dbdataset.what_updated = self.add_what_updated(dbdataset.what_updated, 'script update')
                else:
                    dbdataset.updated_by_script = previous_dbdataset.updated_by_script
                if last_resource_modified <= previous_dbdataset.last_resource_modified:
                    # we keep this so that although we don't normally use it,
                    # we retain the ability to run without touching CKAN
                    dbdataset.last_resource_updated = previous_dbdataset.last_resource_updated
                    dbdataset.last_resource_modified = previous_dbdataset.last_resource_modified
                if latest_of_modifieds < previous_dbdataset.latest_of_modifieds:
                    dbdataset.latest_of_modifieds = previous_dbdataset.latest_of_modifieds
                    if update_frequency is not None and update_frequency > 0:
                        fresh = self.calculate_aging(previous_dbdataset.latest_of_modifieds, update_frequency)
                        dbdataset.fresh = fresh

            self.session.add(dbdataset)

            update_string = '%s, Updated %s' % (self.aging_statuses[fresh], dbdataset.what_updated)
            if dont_hash_script_update or fresh == 0 or update_frequency is None:
                hash_forced = False
                for url, resource_id, force_hash, what_updated in dataset_resources:
                    if force_hash:
                        hash_forced = True
                        resources_to_check.append((url, resource_id, True, what_updated))
                    else:
                        dict_of_lists_add(self.resource_what_updated, what_updated, resource_id)
                if hash_forced:
                    datasets_to_check[dataset_id] = update_string
                else:
                    dict_of_lists_add(self.dataset_what_updated, update_string, dataset_id)
            else:
                datasets_to_check[dataset_id] = update_string
                resources_to_check += dataset_resources
        self.session.commit()
        return datasets_to_check, resources_to_check

    def process_resources(self, dataset_id, previous_dbdataset, resources, dont_hash_script_update,
                          forced_hash_ids=None):
        last_resource_updated = None
        last_resource_modified = None
        dataset_resources = list()
        for resource in resources:
            resource_id = resource['id']
            dict_of_lists_add(self.resource_what_updated, 'total', resource_id)
            url = resource['url']
            name = resource['name']
            revision_last_updated = parser.parse(resource['revision_last_updated'], ignoretz=True)
            last_modified = parser.parse(resource['last_modified'], ignoretz=True)
            if last_resource_modified:
                if last_modified > last_resource_modified:
                    last_resource_updated = resource_id
                    last_resource_modified = last_modified
            else:
                last_resource_updated = resource_id
                last_resource_modified = last_modified
            dbresource = DBResource(run_number=self.run_number, id=resource_id, name=name,
                                    dataset_id=dataset_id, url=url, last_modified=last_modified,
                                    revision_last_updated=revision_last_updated, latest_of_modifieds=last_modified,
                                    what_updated='firstrun')
            if previous_dbdataset is not None:
                try:
                    previous_dbresource = self.session.query(DBResource).filter_by(id=resource_id,
                                                                                   run_number=
                                                                                   previous_dbdataset.run_number).one()
                    if last_modified > previous_dbresource.last_modified:
                        dbresource.what_updated = 'filestore'
                    else:
                        dbresource.last_modified = previous_dbresource.last_modified
                        dbresource.what_updated = 'nothing'
                    if last_modified <= previous_dbresource.latest_of_modifieds:
                        dbresource.latest_of_modifieds = previous_dbresource.latest_of_modifieds
                    dbresource.http_last_modified = previous_dbresource.http_last_modified
                    dbresource.md5_hash = previous_dbresource.md5_hash
                    dbresource.hash_last_modified = previous_dbresource.hash_last_modified
                    dbresource.when_checked = previous_dbresource.when_checked

                except NoResultFound:
                    pass
            self.session.add(dbresource)

            if self.url_internal in url:
                self.internal_what_updated(dbresource, 'internal')
                dict_of_lists_add(self.resource_what_updated, dbresource.what_updated, dbresource.id)
                continue
            if dont_hash_script_update:
                forced_hash = False
            else:
                if forced_hash_ids:
                    forced_hash = resource_id in forced_hash_ids
                else:
                    forced_hash = self.urls_to_check_count < self.no_urls_to_check \
                                  and (dbresource.when_checked is None or
                                       self.now - dbresource.when_checked > datetime.timedelta(days=30))
            if forced_hash:
                dataset_resources.append((url, resource_id, True, dbresource.what_updated))
                self.urls_to_check_count += 1
            else:
                dataset_resources.append((url, resource_id, False, dbresource.what_updated))
        return dataset_resources, last_resource_updated, last_resource_modified

    @staticmethod
    def internal_what_updated(dbresource, url_substr):
        what_updated = '%s-%s' % (url_substr, dbresource.what_updated)
        dbresource.what_updated = what_updated

    def check_urls(self, resources_to_check, user_agent, results=None, hash_results=None):
        def get_domain(x):
            return urlparse(x[0]).netloc
        if results is None:  # pragma: no cover
            resources_to_check = list_distribute_contents(resources_to_check, get_domain)
            results = retrieve(resources_to_check, user_agent)
            if self.testsession:
                serialize_results(self.testsession, results)

        hash_check = list()
        for resource_id in results:
            url, err, http_last_modified, hash, force_hash = results[resource_id]
            if hash:
                dbresource = self.session.query(DBResource).filter_by(id=resource_id,
                                                                      run_number=self.run_number).one()
                if dbresource.md5_hash != hash:  # File changed
                    hash_check.append((url, resource_id, force_hash))

        if hash_results is None:  # pragma: no cover
            hash_check = list_distribute_contents(hash_check, get_domain)
            hash_results = retrieve(hash_check, user_agent)
            if self.testsession:
                serialize_hashresults(self.testsession, hash_results)

        return results, hash_results

    def process_results(self, results, hash_results, resourcecls=Resource):
        datasets_latest_of_modifieds = dict()
        for resource_id in sorted(results):
            url, err, http_last_modified, hash, force_hash = results[resource_id]
            dbresource = self.session.query(DBResource).filter_by(id=resource_id,
                                                                  run_number=self.run_number).one()
            dataset_id = dbresource.dataset_id
            datasetinfo = datasets_latest_of_modifieds.get(dataset_id, dict())
            what_updated = dbresource.what_updated
            update_last_modified = False
            if http_last_modified:
                if dbresource.http_last_modified is None or http_last_modified > dbresource.http_last_modified:
                    dbresource.http_last_modified = http_last_modified
            if hash:
                dbresource.when_checked = self.now
                if dbresource.md5_hash == hash:  # File unchanged
                    what_updated = self.add_what_updated(what_updated, 'same hash')
                else:  # File updated
                    hash_to_set = hash
                    hash_url, hash_err, hash_http_last_modified, hash_hash, force_hash = hash_results[resource_id]
                    if hash_http_last_modified:
                        if dbresource.http_last_modified is None or hash_http_last_modified > dbresource.http_last_modified:
                            dbresource.http_last_modified = hash_http_last_modified
                    if hash_hash:
                        if hash_hash == hash:
                            if dbresource.md5_hash is None:  # First occurrence of resource eg. first run - don't use hash
                                # for last modified field (and hence freshness calculation)
                                dbresource.what_updated = self.add_what_updated(what_updated,
                                                                                'first hash')
                                what_updated = dbresource.what_updated
                            else:
                                # Check if hash has occurred before
                                # select distinct md5_hash from dbresources where id = '714ef7b5-a303-4e4f-be2f-03b2ce2933c7' and md5_hash='2f3cd6a6fce5ad4d7001780846ad87a7';
                                if self.session.query(exists().where(
                                        and_(DBResource.id == resource_id, DBResource.md5_hash == hash))).scalar():
                                    dbresource.what_updated = self.add_what_updated(what_updated, 'repeat hash')
                                    what_updated = dbresource.what_updated
                                else:
                                    what_updated, _ = self.set_latest_of_modifieds(dbresource, self.now, 'hash')
                                    dbresource.hash_last_modified = self.now
                                    update_last_modified = True
                            dbresource.api = False
                        else:
                            hash_to_set = hash_hash
                            what_updated = self.add_what_updated(what_updated, 'api')
                            dbresource.api = True
                    if hash_err:
                        what_updated = self.add_what_updated(what_updated, 'error')
                        dbresource.error = hash_err
                    dbresource.md5_hash = hash_to_set
            if err:
                dbresource.when_checked = self.now
                what_updated = self.add_what_updated(what_updated, 'error')
                dbresource.error = err
            datasetinfo[resource_id] = (dbresource.error, dbresource.latest_of_modifieds, dbresource.what_updated)
            datasets_latest_of_modifieds[dataset_id] = datasetinfo
            dict_of_lists_add(self.resource_what_updated, what_updated, resource_id)
            if update_last_modified and self.do_touch:
                try:
                    logger.info('Updating last modified for resource %s' % resource_id)
                    resource = resourcecls.read_from_hdx(resource_id)
                    if resource:
                        last_modified = parser.parse(resource['last_modified'])
                        dbdataset = self.session.query(DBDataset).filter_by(id=dataset_id,
                                                                            run_number=self.run_number).one()
                        update_frequency = dbdataset.update_frequency
                        if update_frequency > 0:
                            if self.calculate_aging(last_modified, update_frequency) == 0:
                                dotouch = False
                            else:
                                dotouch = True
                        else:
                            dotouch = True
                        if dotouch:
                            self.resource_last_modified_count += 1
                            logger.info('Resource last modified count: %d' % self.resource_last_modified_count)
                            resource['last_modified'] = dbresource.latest_of_modifieds.isoformat()
                            resource.update_in_hdx(operation='patch', batch_mode='KEEP_OLD', skip_validation=True,
                                                   ignore_check=True)
                        else:
                            logger.info("Didn't update last modified for resource %s as it is fresh!" % resource_id)
                    else:
                        logger.error('Last modified update failed for id %s! Resource does not exist.' % resource_id)
                except HDXError:
                    logger.exception('Last modified update failed for id %s!' % resource_id)
        self.session.commit()
        return datasets_latest_of_modifieds

    def update_dataset_latest_of_modifieds(self, datasets_to_check, datasets_latest_of_modifieds):
        for dataset_id in datasets_latest_of_modifieds:
            dbdataset = self.session.query(DBDataset).filter_by(id=dataset_id,
                                                                run_number=self.run_number).one()
            dataset = datasets_latest_of_modifieds[dataset_id]
            dataset_latest_of_modifieds = dbdataset.latest_of_modifieds
            dataset_what_updated = dbdataset.what_updated
            last_resource_modified = dbdataset.last_resource_modified
            last_resource_updated = dbdataset.last_resource_updated
            all_errors = True
            for resource_id in sorted(dataset):
                err, new_last_resource_modified, new_last_resource_what_updated = dataset[resource_id]
                if not err:
                    all_errors = False
                if new_last_resource_modified:
                    if new_last_resource_modified > last_resource_modified:
                        last_resource_updated = resource_id
                        last_resource_modified = new_last_resource_modified
                    if new_last_resource_modified > dataset_latest_of_modifieds:
                        dataset_latest_of_modifieds = new_last_resource_modified
                        dataset_what_updated = new_last_resource_what_updated
            dbdataset.last_resource_updated = last_resource_updated
            dbdataset.last_resource_modified = last_resource_modified
            self.set_latest_of_modifieds(dbdataset, dataset_latest_of_modifieds, dataset_what_updated)
            update_frequency = dbdataset.update_frequency
            if update_frequency is not None and update_frequency > 0:
                dbdataset.fresh = self.calculate_aging(dbdataset.latest_of_modifieds, update_frequency)
            dbdataset.error = all_errors
            status = '%s, Updated %s' % (self.aging_statuses[dbdataset.fresh], dbdataset.what_updated)
            if all_errors:
                status = '%s,error' % status
            dict_of_lists_add(self.dataset_what_updated, status, dataset_id)
        self.session.commit()
        for dataset_id in datasets_to_check:
            if dataset_id in datasets_latest_of_modifieds:
                continue
            dict_of_lists_add(self.dataset_what_updated, datasets_to_check[dataset_id], dataset_id)

    def output_counts(self):
        def add_what_updated_str(hdxobject_what_updated):
            nonlocal output_str
            output_str += '\n* total: %d *' % len(hdxobject_what_updated['total'])
            for countstr in sorted(hdxobject_what_updated):
                if countstr != 'total':
                    output_str += ',\n%s: %d' % (countstr, len(hdxobject_what_updated[countstr]))

        output_str = '\n*** Resources ***'
        add_what_updated_str(self.resource_what_updated)
        output_str += '\n\n*** Datasets ***'
        add_what_updated_str(self.dataset_what_updated)
        output_str += '\n\n%d datasets have update frequency of Live' % self.live_update
        output_str += '\n%d datasets have update frequency of Never' % self.never_update
        output_str += '\n%d datasets have update frequency of Adhoc' % self.adhoc_update

        logger.info(output_str)
        return output_str

    @staticmethod
    def set_latest_of_modifieds(dbobject, modified_date, what_updated):
        if modified_date > dbobject.latest_of_modifieds:
            dbobject.latest_of_modifieds = modified_date
            dbobject.what_updated = DataFreshness.add_what_updated(dbobject.what_updated, what_updated)
            update = True
        else:
            update = False
        return dbobject.what_updated, update

    @staticmethod
    def add_what_updated(prev_what_updated, what_updated):
        if what_updated in prev_what_updated:
            return prev_what_updated
        if prev_what_updated != 'nothing' and prev_what_updated != 'firstrun':
            if what_updated != 'nothing':
                return '%s,%s' % (prev_what_updated, what_updated)
            return prev_what_updated
        else:
            return what_updated

    def calculate_aging(self, last_modified, update_frequency):
        delta = self.now - last_modified
        if delta >= self.aging[update_frequency]['Delinquent']:
            return 3
        elif delta >= self.aging[update_frequency]['Overdue']:
            return 2
        elif delta >= self.aging[update_frequency]['Due']:
            return 1
        return 0





# -*- coding: utf-8 -*-
'''
Serialize:
---------

Serialize datasets and other types

'''
from collections import OrderedDict

from hdx.data.dataset import Dataset

from hdx.freshness.testdata.dbtestdataset import DBTestDataset
from hdx.freshness.testdata.dbtestdate import DBTestDate
from hdx.freshness.testdata.dbtesthashresult import DBTestHashResult
from hdx.freshness.testdata.dbtestresource import DBTestResource
from hdx.freshness.testdata.dbtestresult import DBTestResult


def serialize_datasets(session, datasets):
    for dataset in datasets:
        dataset_id = dataset['id']
        dbtestdataset = DBTestDataset(id=dataset_id, organization_id=dataset['organization']['id'],
                                      organization_name=dataset['organization']['name'],
                                      organization_title=dataset['organization']['title'], dataset_name=dataset['name'],
                                      dataset_title=dataset['title'], dataset_private=dataset['private'],
                                      dataset_maintainer=dataset['maintainer'],
                                      dataset_date=dataset.get('dataset_date'),
                                      update_frequency=dataset.get('data_update_frequency'),
                                      review_date=dataset['review_date'],
                                      last_modified=dataset['last_modified'],
                                      updated_by_script=dataset.get('updated_by_script'),
                                      metadata_modified=dataset['metadata_modified'],
                                      is_requestdata_type=dataset.get('is_requestdata_type'),
                                      dataset_location=','.join([x['name'] for x in dataset['groups']]))
        session.add(dbtestdataset)
        for resource in dataset.get_resources():
            dbtestresource = DBTestResource(id=resource['id'], name=resource['name'], dataset_id=dataset_id,
                                            format=resource['format'], url=resource['url'],
                                            revision_last_updated=resource['revision_last_updated'],
                                            last_modified=resource['last_modified'])
            session.add(dbtestresource)
    session.commit()


def serialize_now(session, now):
    dbtestdate = DBTestDate(test_date=now)
    session.add(dbtestdate)
    session.commit()


def serialize_results(session, results):
    for id in results:
        url, err, http_last_modified, hash, force_hash = results[id]
        dbtestresult = DBTestResult(id=id, url=url, err=err, http_last_modified=http_last_modified,
                                    hash=hash, force_hash=force_hash)
        session.add(dbtestresult)
    session.commit()


def serialize_hashresults(session, hash_results):
    for id in hash_results:
        url, err, http_last_modified, hash, force_hash = hash_results[id]
        dbtesthashresult = DBTestHashResult(id=id, url=url, err=err, http_last_modified=http_last_modified,
                                            hash=hash, force_hash=force_hash)
        session.add(dbtesthashresult)
    session.commit()


def deserialize_datasets(session):
    datasets = OrderedDict()
    for dbtestdataset in session.query(DBTestDataset):
        dataset_id = dbtestdataset.id
        organization = {
            'id': dbtestdataset.organization_id,
            'name': dbtestdataset.organization_name,
            'title': dbtestdataset.organization_title
        }
        dataset = Dataset({
            'id': dataset_id,
            'organization': organization,
            'name': dbtestdataset.dataset_name,
            'title': dbtestdataset.dataset_title,
            'private': dbtestdataset.dataset_private,
            'maintainer': dbtestdataset.dataset_maintainer,
            'dataset_date': dbtestdataset.dataset_date,
            'data_update_frequency': dbtestdataset.update_frequency,
            'review_date': dbtestdataset.review_date,
            'last_modified': dbtestdataset.last_modified,
            'updated_by_script': dbtestdataset.updated_by_script,
            'metadata_modified': dbtestdataset.metadata_modified,
            'groups': [{'name': x} for x in dbtestdataset.dataset_location.split(',')]
        })
        dataset.set_requestable(dbtestdataset.is_requestdata_type)
        datasets[dataset_id] = dataset
    for dbtestresource in session.query(DBTestResource):
        dataset = datasets[dbtestresource.dataset_id]
        resource = {
            'id': dbtestresource.id,
            'name': dbtestresource.name,
            'format': dbtestresource.format,
            'url': dbtestresource.url,
            'revision_last_updated': dbtestresource.revision_last_updated,
            'last_modified': dbtestresource.last_modified
        }
        dataset.get_resources().append(resource)
    return datasets.values()


def deserialize_now(session):
    return session.query(DBTestDate.test_date).scalar()


def deserialize_results(session):
    results = dict()
    for dbtestresult in session.query(DBTestResult):
        results[dbtestresult.id] = (dbtestresult.url, dbtestresult.err, dbtestresult.http_last_modified,
                                    dbtestresult.hash, dbtestresult.force_hash)
    return results


def deserialize_hashresults(session):
    hash_results = dict()
    for dbtesthashresult in session.query(DBTestHashResult):
        hash_results[dbtesthashresult.id] = (dbtesthashresult.url, dbtesthashresult.err,
                                             dbtesthashresult.http_last_modified, dbtesthashresult.hash,
                                             dbtesthashresult.force_hash)
    return hash_results



#!/usr/bin/python
# -*- coding: utf-8 -*-
'''
Retrieval
---------

Retrieve urls and categorise them

'''
import asyncio
import hashlib
import logging
from timeit import default_timer as timer

import aiohttp
import tqdm
import uvloop
from dateutil import parser

from hdx.freshness import retry
from hdx.freshness.ratelimiter import RateLimiter

logger = logging.getLogger(__name__)


async def fetch(metadata, session):
    url = metadata[0]
    resource_id = metadata[1]
    force_hash = metadata[2]

    async def fn(response):
        last_modified_str = response.headers.get('Last-Modified')
        http_last_modified = None
        if last_modified_str:
            try:
                http_last_modified = parser.parse(last_modified_str, ignoretz=True)
                if not force_hash:
                    response.close()
                    return resource_id, url, None, http_last_modified, None, False
            except (ValueError, OverflowError):
                pass
        length = response.headers.get('Content-Length')
        if length and int(length) > 419430400:
            response.close()
            err = 'File too large to hash!'
            return resource_id, url, err, http_last_modified, None, force_hash
        logger.info('Hashing %s' % url)
        try:
            md5hash = hashlib.md5()
            async for chunk in response.content.iter_chunked(10240):
                if chunk:
                    md5hash.update(chunk)
            return resource_id, url, None, http_last_modified, md5hash.hexdigest(), force_hash
        except Exception as exc:
            try:
                code = exc.code
            except AttributeError:
                code = ''
            err = 'Exception during hashing: code=%s message=%s raised=%s.%s url=%s' % (code, exc,
                                                                                         exc.__class__.__module__,
                                                                                         exc.__class__.__qualname__,
                                                                                         url)
            if http_last_modified:
                return resource_id, url, err, http_last_modified, None, force_hash
            raise aiohttp.ClientResponseError(code=code, message=err,
                                              request_info=response.request_info, history=response.history) from exc

    try:
        return await retry.send_http(session, 'get', url,
                                     retries=2,
                                     interval=5,
                                     backoff=4,
                                     fn=fn)
    except Exception as e:
        return resource_id, url, str(e), None, None, force_hash


async def check_urls(urls, loop, user_agent):
    tasks = list()

    conn = aiohttp.TCPConnector(limit=100, limit_per_host=1, loop=loop)
    timeout = aiohttp.ClientTimeout(total=60 * 60, sock_connect=30, sock_read=30)
    async with aiohttp.ClientSession(connector=conn, timeout=timeout, loop=loop,
                                     headers={'User-Agent': user_agent}) as session:
        session = RateLimiter(session)
        for metadata in urls:
            task = fetch(metadata, session)
            tasks.append(task)
        responses = dict()
        for f in tqdm.tqdm(asyncio.as_completed(tasks), total=len(tasks)):
            resource_id, url, err, http_last_modified, hash, force_hash = await f
            responses[resource_id] = (url, err, http_last_modified, hash, force_hash)
        return responses


def retrieve(urls, user_agent):
    start_time = timer()
    loop = uvloop.new_event_loop()
    asyncio.set_event_loop(loop)
    future = asyncio.ensure_future(check_urls(urls, loop, user_agent))
    results = loop.run_until_complete(future)
    logger.info('Execution time: %s seconds' % (timer() - start_time))
    loop.run_until_complete(asyncio.sleep(0.250))
    loop.close()
    return results



# -*- coding: utf-8 -*-
"""Version utility"""
import logging

from hdx.utilities.path import script_dir_plus_file

logger = logging.getLogger(__name__)


def get_freshness_version():
    version_file = open(script_dir_plus_file('version.txt', get_freshness_version))
    return version_file.read().strip()



#!/usr/bin/python
# -*- coding: utf-8 -*-
'''
Rate limit
----------

Limit connections per timeframe to host

'''
import asyncio
import time
from urllib.parse import urlsplit


class RateLimiter:
    RATE = 9 / 60
    MAX_TOKENS = 10

    def __init__(self, session):
        self.session = session
        self.start_time = time.monotonic()
        self.tokens = dict()

    async def get(self, url, *args, **kwargs):
        host = urlsplit(url).netloc
        await self.wait_for_token(host)
        return self.session.get(url, *args, **kwargs)

    async def wait_for_token(self, host):
        if host not in self.tokens:
            self.tokens[host] = [self.MAX_TOKENS, self.start_time]
        while self.tokens[host][0] < 1:
            self.add_new_tokens(host)
            await asyncio.sleep(1)
        self.tokens[host][0] -= 1

    def add_new_tokens(self, host):
        now = time.monotonic()
        time_since_update = now - self.tokens[host][1]
        new_tokens = time_since_update * self.RATE
        if new_tokens > 1:
            self.tokens[host][0] = min(self.tokens[host][0] + new_tokens, self.MAX_TOKENS)
            self.tokens[host][1] = now



import asyncio
import logging

import aiohttp

logger = logging.getLogger(__name__)


HTTP_STATUS_CODES_TO_RETRY = [500, 502, 503, 504]


class FailedRequest(Exception):
    """
    A wrapper of all possible exception during a HTTP request
    """
    code = 0
    message = ''
    url = ''
    raised = ''

    def __init__(self, *, raised='', message='', code='', url=''):
        self.raised = raised
        self.message = message
        self.code = code
        self.url = url

        super().__init__('code={c} message={m} raised={r} url={u}'.format(
            c=self.code, m=self.message, r=self.raised, u=self.url))


async def send_http(session, method, url, *,
                    retries=1,
                    interval=1,
                    backoff=2,
                    http_status_codes_to_retry=HTTP_STATUS_CODES_TO_RETRY,
                    fn=lambda x:x,
                    **kwargs):
    """
    Sends a HTTP request and implements a retry logic.

    Arguments:
        session (obj): A client aiohttp session object
        method (str): Method to use
        url (str): URL for the request
        retries (int): Number of times to retry in case of failure
        interval (float): Time to wait before retries
        backoff (int): Multiply interval by this factor after each failure
        http_status_codes_to_retry (List[int]): List of status codes to retry
        fn (Callable[[x],x]: Function to call on successful connection
    """
    backoff_interval = interval
    raised_exc = None
    attempt = 0

    if method not in ['get', 'patch', 'post']:
        raise ValueError

    if retries == -1:  # -1 means retry indefinitely
        attempt = -1
    elif retries == 0:  # Zero means don't retry
        attempt = 1
    else:  # any other value means retry N times
        attempt = retries + 1

    while attempt != 0:
        if raised_exc:
            logger.error('Caught "%s" url:%s method:%s, remaining tries %s, '
                    'sleeping %.2fsecs', raised_exc, method.upper(), url,
                    attempt, backoff_interval)
            await asyncio.sleep(backoff_interval)
            # bump interval for the next possible attempt
            backoff_interval *= backoff
        # logger.info('sending %s %s with %s', method.upper(), url, kwargs)
        try:
            async with await getattr(session, method)(url, **kwargs) as response:
                if response.status == 200:
                    return await fn(response)
                elif response.status in http_status_codes_to_retry:
                    logger.error(
                        'Received invalid response code:%s error:%s'
                        ' response:%s url:%s', response.status, '', response.reason, url)
                    raise aiohttp.ClientResponseError(
                        code=response.status, message=response.reason, request_info=response.request_info,
                        history=response.history)
                else:
                    raise FailedRequest(
                        code=response.status, message='Non-retryable response code',
                        raised='aiohttp.ClientResponseError', url=url)
        except aiohttp.ClientError as exc:
            try:
                code = exc.code
            except AttributeError:
                code = ''
            raised_exc = FailedRequest(code=code, message=exc,
                                       raised='%s.%s' % (exc.__class__.__module__, exc.__class__.__qualname__), url=url)
        except asyncio.TimeoutError as exc:
            raised_exc = FailedRequest(code='', message='asyncio.TimeoutError',
                                       raised='%s.%s' % (exc.__class__.__module__, exc.__class__.__qualname__), url=url)
        else:
            raised_exc = None
            break

        attempt -= 1

    if raised_exc:
        raise raised_exc


# -*- coding: utf-8 -*-
from hdx.utilities.database import Base
from sqlalchemy import Boolean
from sqlalchemy import Column, DateTime, String
from sqlalchemy import ForeignKey
from sqlalchemy import Integer

from hdx.freshness.database.dbinfodataset import DBInfoDataset
from hdx.freshness.database.dbrun import DBRun


class DBResource(Base):
    run_number = Column(Integer, ForeignKey(DBRun.run_number), primary_key=True)
    id = Column(String, primary_key=True)
    name = Column(String, nullable=False)
    dataset_id = Column(String, ForeignKey(DBInfoDataset.id), nullable=False)
    url = Column(String, nullable=False)
    last_modified = Column(DateTime, nullable=False)
    revision_last_updated = Column(DateTime, default=None)  # this field and above are CKAN fields
    latest_of_modifieds = Column(DateTime, nullable=False)
    what_updated = Column(String, nullable=False)
    http_last_modified = Column(DateTime, default=None)
    md5_hash = Column(String, default=None)
    hash_last_modified = Column(DateTime, default=None)
    when_checked = Column(DateTime, default=None)
    api = Column(Boolean)
    error = Column(String)

    def __repr__(self):
        output = '<Resource(run number=%d, id=%s, name=%s, ' % (self.run_number, self.id, self.name)
        output += 'dataset id=%s,\nurl=%s,\n' % (self.dataset_id, self.url)
        output += 'last modified=%s, revision last updated=%s,\n' % (str(self.last_modified),
                                                                     str(self.revision_last_updated))
        output += 'latest of modifieds=%s, what updated=%s,\n' % (str(self.latest_of_modifieds), str(self.what_updated))
        output += 'http last modified=%s,\n' % str(self.http_last_modified)
        output += 'MD5 hash=%s, hash last modified=%s, ' % (self.md5_hash, str(self.hash_last_modified))
        output += 'when checked=%s,\n' % str(self.when_checked)
        output += 'api=%s, error=%s)>' % (str(self.api), str(self.error))
        return output
